import 'dart:async';

/// A class contains sinks that will be closed
/// when [dispose] get called
class StreamsDisposeBag implements Future<void> {
  StreamsDisposeBag(List disposables)
      : _disposables = disposables
                .map(
                  (o) => _StreamsDisposable(o),
                )
                .toList() ??
            [];

  //
  List<_StreamsDisposable> _disposables;

  /// A function that add a new [Sink] into the list
  ///
  /// The sink that already added will be skipped
  void add(disposable) {
    if (_disposables.map((o) => o.object).contains(disposable)) return;
    _disposables.add(_StreamsDisposable(disposable));
  }

  // A function that will close all [Sink] that added
  Future<void> dispose() {
    return Future.wait(_disposables.map((o) => o.dispose()));
  }

  @override
  Stream asStream() {
    return dispose().asStream();
  }

  @override
  Future catchError(Function onError, {bool Function(Object error) test}) {
    return dispose().catchError(onError, test: test);
  }

  @override
  Future<R> then<R>(FutureOr<R> Function(dynamic) onValue, {Function onError}) {
    return dispose().then(onValue, onError: onError);
  }

  @override
  Future timeout(Duration timeLimit, {FutureOr Function() onTimeout}) {
    return dispose().timeout(timeLimit, onTimeout: onTimeout);
  }

  @override
  Future whenComplete(FutureOr Function() action) {
    return dispose().whenComplete(action);
  }
}

class _StreamsDisposable {
  _StreamsDisposable(this.object)
      : assert(object == null || object is Sink || object is StreamSubscription || object is Timer);

  //
  final dynamic object;

  Future<void> dispose() async {
    // Sink
    if (object is Sink) {
      (object as Sink).close();
    }

    // Stream Subscription
    if (object is StreamSubscription) {
      (object as StreamSubscription).cancel();
    }

    // Timer
    if (object is Timer) {
      (object as Timer).cancel();
    }
  }
}
